import itertools
import string
from io import StringIO
from pathlib import Path
from typing import List

import click
import inflect as ifl
import pyperclip
from InquirerPy.base import Choice

from ignorer import prompts

inflect = ifl.engine()

multiselect_controls = ("↑/↓: Move up and down\n"
                        "Control + Z: Toggle selection\n"
                        "Control + A: Toggle all on/off\n"
                        "Enter: Confirm\n")


class Underline:
    end = '\033[0m'
    underline = '\033[4m'


def generate_gitignore(templates: List[Path], remove_duplicates: bool, remove_comments: bool) -> str:
    gitignore = StringIO()

    for template in templates:
        for line in template.open().readlines():
            if remove_duplicates and line in gitignore.getvalue() and line not in string.whitespace:
                continue

            if remove_comments and (line.startswith("#") or line in string.whitespace):
                continue

            gitignore.write(line)

        if template != templates[-1] and not remove_comments:
            gitignore.write("\n")

    gitignore.seek(0)

    return gitignore.read()


def save_gitignore(gitignore: str, use_cwd: bool = False) -> None:
    while True:
        if use_cwd:
            save_path = Path.cwd() / ".gitignore"
        else:
            def path_postprocessor(path: str) -> str:
                path = Path(path).joinpath(".gitignore")

                path.parent.mkdir(parents=True, exist_ok=True)
                return path.expanduser().resolve()

            save_path = prompts.filepath(
                message="Where should your .gitignore be saved?",
                instruction="Enter a path.",
                long_instruction="Both absolute and relative paths are supported.\n"
                                 "Nonexistent directories will be created as necessary.",
                multicolumn_complete=True,
                only_directories=True,
                validate=lambda path: bool(path),
                invalid_message="You must enter a path.",
                filter=path_postprocessor,
                transformer=path_postprocessor).execute()

        if save_path.exists():
            overwrite_choices = [
                Choice(name="Overwrite it", value=True),
                Choice(name="Choose another place", value=False),
            ]

            overwrite = prompts.select(
                message="A .gitignore file already exists there. What do you wanna do?",
                choices=overwrite_choices).execute()

            if overwrite:
                save_path.write_text(gitignore)
                break
            else:
                use_cwd = False
                continue

        else:
            save_path.parent.mkdir(parents=True, exist_ok=True)
            save_path.write_text(gitignore)
            break

    print(f"\n.gitignore saved to {save_path}.")


def get_templates() -> List[Path]:
    categories = [
        Choice(name="Languages and frameworks", value=["*.gitignore", "community/**/*.gitignore"]),
        Choice(name="Editors and operating systems", value=["Global/*.gitignore"]),
        Choice(name="Just show me everything", value=[]),
    ]

    category = (prompts.select(message="Choose a category.", choices=categories).execute() or
                itertools.chain.from_iterable([choices.value for choices in categories]))

    repo_path = Path(__file__).parent / "gitignore"
    templates: List[Path] = []

    for template in itertools.chain(*[repo_path.glob(pattern) for pattern in category]):
        name = template.stem

        if template.parent.name == "Global" and any(t.name == template.stem for t in templates):
            name += " (Global)"

        templates.append(Choice(name=name, value=template))

    templates.sort(key=lambda t: t.name.lower())

    return prompts.fuzzy(
        message="Choose some templates.",
        instruction="You can type to search for specific ones.",
        long_instruction=multiselect_controls,
        choices=templates,
        multiselect=True,
        transformer=lambda result: f"Generating from {inflect.no('template', len(result))}."
    ).execute()


def get_modifiers() -> List[str]:
    modifier_choices = [
        Choice(name="Generate without duplicates", value="no_duplicates"),
        Choice(name="Generate without comments and empty lines", value="no_comments"),
    ]

    def modifier_transformer(result: List[str]) -> str:
        modifier_transformations = {
            "Generate without duplicates": ["duplicates"],
            "Generate without comments and empty lines": ["comments", "empty lines"],
        }

        words = [x for sublist in [modifier_transformations[r] for r in result] for x in sublist]
        return f"Generating without {inflect.join(words, conj='or')}."

    modifiers = prompts.select(message="Choose some options.",
                               instruction="Or skip this question with Control + D.",
                               long_instruction=multiselect_controls,
                               choices=modifier_choices,
                               multiselect=True,
                               mandatory=False,
                               transformer=lambda result: modifier_transformer(result)).execute() or []

    return modifiers


def get_output() -> str:
    output_choices = [
        Choice(name="Save to current directory", value="cwd"),
        Choice(name="Save to another directory", value="elsewhere"),
        Choice(name="Print to stdout", value="stdout"),
        Choice(name="Copy to clipboard", value="clipboard"),
    ]

    output = prompts.select(message="Choose how to save your .gitignore.",
                            choices=output_choices).execute()

    return output


def main():
    print("ignorer © 2022-present celsius narhwal. Licensed under MIT (see --license).\n")
    templates = get_templates()
    modifiers = get_modifiers()
    output = get_output()

    remove_duplicates, remove_comments = "no_duplicates" in modifiers, "no_comments" in modifiers

    gitignore = generate_gitignore(templates, remove_duplicates, remove_comments)

    if output == "cwd":
        save_gitignore(gitignore, use_cwd=True)
    elif output == "elsewhere":
        save_gitignore(gitignore)
    elif output == "stdout":
        print(f"\n{Underline.underline}Generated .gitignore{Underline.end}\n{gitignore}")
    elif output == "clipboard":
        pyperclip.copy(gitignore)
        print("\n.gitignore copied to clipboard.")


@click.command()
@click.option("--license", "show_license", is_flag=True, help="See ignorer's license.")
def cli(show_license: bool):
    if show_license:
        print((Path(__file__).parent / "LICENSE").read_text())
    else:
        while True:
            try:
                main()
                break
            except (KeyboardInterrupt, EOFError):
                print("\nExiting.")
                break


if __name__ == '__main__':
    cli()



from InquirerPy import inquirer


def prompt_decorator(func):
    def wrapper(*args, **kwargs):
        if kwargs.get("multiselect"):
            keybindings = {
                "toggle": [
                    {"key": "c-z"}
                ],
                "toggle-all-true": [
                    {"key": "c-a"}
                ],
                "toggle-all-false": [
                    {"key": "c-a"}
                ],
            }
        else:
            keybindings = {
                "answer": [
                    {"key": "enter"},
                    {"key": "c-z"}
                ],
            }

        keybindings["skip"] = [{"key": "c-d"}]

        kwargs["keybindings"] = keybindings
        kwargs["qmark"] = "•"
        kwargs["amark"] = "✓"
        kwargs["mandatory_message"] = "You can't skip this."

        return func(*args, **kwargs)

    return wrapper


@prompt_decorator
def checkbox(*args, **kwargs):
    return inquirer.checkbox(*args, **kwargs)


@prompt_decorator
def confirm(*args, **kwargs):
    return inquirer.confirm(*args, **kwargs)


@prompt_decorator
def expand(*args, **kwargs):
    return inquirer.expand(*args, **kwargs)


@prompt_decorator
def filepath(*args, **kwargs):
    return inquirer.filepath(*args, **kwargs)


@prompt_decorator
def fuzzy(*args, **kwargs):
    return inquirer.fuzzy(*args, **kwargs)


@prompt_decorator
def text(*args, **kwargs):
    return inquirer.text(*args, **kwargs)


@prompt_decorator
def select(*args, **kwargs):
    return inquirer.select(*args, **kwargs)


@prompt_decorator
def number(*args, **kwargs):
    return inquirer.number(*args, **kwargs)


@prompt_decorator
def rawlist(*args, **kwargs):
    return inquirer.rawlist(*args, **kwargs)


@prompt_decorator
def secret(*args, **kwargs):
    return inquirer.secret(*args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ignorer']

package_data = \
{'': ['*'],
 'ignorer': ['gitignore/*',
             'gitignore/.github/*',
             'gitignore/Global/*',
             'gitignore/community/*',
             'gitignore/community/AWS/*',
             'gitignore/community/DotNet/*',
             'gitignore/community/Elixir/*',
             'gitignore/community/GNOME/*',
             'gitignore/community/Golang/*',
             'gitignore/community/Java/*',
             'gitignore/community/JavaScript/*',
             'gitignore/community/Linux/*',
             'gitignore/community/PHP/*',
             'gitignore/community/Python/*',
             'gitignore/community/embedded/*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'inflect>=6.0.2,<7.0.0',
 'inquirerpy>=0.3.4,<0.4.0',
 'pyperclip>=1.8.2,<2.0.0']

entry_points = \
{'console_scripts': ['ignorer = ignorer.ignorer:cli']}

setup_kwargs = {
    'name': 'ignorer',
    'version': '1.0.2',
    'description': 'Generate .gitignore files from your command line',
    'long_description': "# ignorer\n\nignorer generates .gitignore files from your command line using\nGitHub's [vast repository](https://github.com/github/gitignore) of .gitignore templates.\n\n## Installation\n\n### pipx (recommended)\n\nThe recommended, cross-platform, way of installing ignorer is via [pipx](https://pypa.github.io/pipx/).\n\n```bash\npipx install ignorer\n```\n\n### Homebrew (macOS and Linux)\n\nA [Homebrew](https://brew.sh/) formula for ignorer is available from\nthe [Houkago Tea Tap](https://github.com/celsiusnarhwal/homebrew-htt).\n\n```bash\nbrew tap celsiusnarhwal/htt\nbrew install ignorer\n```\n\n### pip (not recommended)\n\nignorer can be installed via pip like any other Python package, but unless you're going to make a virtual environment\nfor it yourself, you're strongly encouraged to use pipx or Homebrew.\n\n```bash\npip install ignorer\n```\n\n## Usage\n\nSimply invoke ignorer from the command line:\n\n```bash\nignorer\n```\n\nThat's it. ignorer will interactively guide you through the rest.\n\nSee it in action below.\n\n![A GIF demonstrating the usage of ignorer.](media/demonstration.gif)\n\n## Acknowledgements\n\nignorer's functionality is largely based on JetBrains' [.ignore plugin](https://github.com/JetBrains/idea-gitignore)\nfor their IDEs.\n\n## License\n\nignorer is licensed under the [MIT License](LICENSE.md).\n",
    'author': 'celsius narhwal',
    'author_email': 'hello@celsiusnarhwal.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/celsiusnarhwal/ignorer',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
